/**
 * State of the `useAsync` hook (Managed by the AsyncReducer).
 */
export interface State<T> {
    status: 'idle' | 'pending' | 'rejected' | 'resolved';
    data: null | T;
    error: null | Error;
}
/**
 * Action passed to the AsyncReducer.
 */
export declare type Action<T> = Partial<State<T>>;
export declare type AsyncReducer<T> = (prevState: State<T>, action: Action<T>) => State<T>;
/**
 * Async hook derived from Kent C Dodd's react-performance workshop:
 * https://github.com/kentcdodds/react-performance/blob/94c95c009d41c4d29258dfc5cd9c9dd4e03a063d/src/utils.js#L22
 *
 * @example
 * const { data, error, status, run } = useAsync<Post>()
 *
 * React.useEffect(() => {
 *   run(async () => {
 *     const response = await fetch(
 *       `https://jsonplaceholder.typicode.com/posts/${props.postId}`
 *     )
 *
 *     const data = await response.json() as Post
 *
 *     if (!response.ok) {
 *       throw new Error(`Failed to fetch post: #${props.postId}.`)
 *     }
 *
 *     return data
 *   })
 * }, [props.postId, run])
 *
 * return status === 'idle' || status === 'pending'
 *   ? <div>loading</div>
 *   : status === 'rejected'
 *   ? <div>Failed to fetch post: ${postId}</div>
 *   : (
 *     <>
 *       <h1>data.title</h1>
 *       <p>data.body</p>
 *     </>
 *   )
 */
export declare function useAsync<T>(initialState?: Action<T>): {
    isIdle: boolean;
    isLoading: boolean;
    isError: boolean;
    isSuccess: boolean;
    setData: (data: T) => void;
    setError: (error: Error) => void;
    error: Error | null;
    status: "idle" | "pending" | "rejected" | "resolved";
    data: T | null;
    run: (promise: Promise<T> | (() => Promise<T>)) => Promise<Error | T>;
    reset: () => void;
};
export default useAsync;
